#!/usr/bin/env python

import logging
import logging.config
import os
import re
import sys
import time

import jwt
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

MEASURED_TIME_LOG = "/tmp/measured_time.log"
SAMPLE_SIZE = 10
LOGGER_FORMAT = '(%(process)d) %(asctime)s %(filename)s (line %(lineno)s) | %(asctime)s %(levelname)-8s %(message)s'


def get_instrument_type_from_name(instrument_name):
    if instrument_name.startswith("M"):
        return "MISEQ"
    if instrument_name.startswith("SN") \
            or instrument_name.startswith("D") \
            or instrument_name.startswith("L") \
            or instrument_name[0].isdigit() \
            or instrument_name.startswith("HSQ"):
        return "HISEQ"
    if instrument_name.startswith("K"):
        return "HISEQ4k"
    if instrument_name.startswith("HW"):
        return "GAIIX"
    if instrument_name.startswith("NS") \
            or instrument_name.startswith("NB") \
            or instrument_name.startswith("NDX"):
        return "NEXTSEQ"
    return None


def get_expected_time(class_name, func_name):
    if not os.path.isfile(MEASURED_TIME_LOG):
        return None
    times = []
    with open(MEASURED_TIME_LOG) as f:
        for line in f.readlines():
            if line.startswith("{} {} ".format(class_name, func_name)):
                times.append(int(line.lstrip("{} {} ".format(class_name, func_name))))
    result = times[-SAMPLE_SIZE:]
    return None if len(result) == 0 else int(round(sum(result) / float(len(result))))


def set_measured_time(class_name, func_name, measured_time):
    with open(MEASURED_TIME_LOG, 'a') as f:
        f.write("{} {} {}\n".format(class_name, func_name, measured_time))


def timeit(func):
    def timed(self, seq_folder):
        class_name = self.__class__.__name__
        func_name = func.__name__
        ts = time.time()
        result = func(self, seq_folder)
        te = time.time()
        measured_time = int(round(te - ts))
        set_measured_time(class_name, func_name, measured_time)
        return result

    return timed


def read_in_chunks(file_object, chunk_size=65536):
    while True:
        data = file_object.read(chunk_size)
        if not data:
            break
        yield data


class _ExcludeErrorsFilter(logging.Filter):
    def filter(self, record):
        """Filters out log messages with log level ERROR (numeric value: 40) or higher."""
        return record.levelno < 40


class Logging(object):
    config = {
        'version': 1,
        'filters': {
            'exclude_errors': {
                '()': _ExcludeErrorsFilter
            }
        },
        'formatters': {
            'include_process': {
                'format': LOGGER_FORMAT
            }
        },
        'handlers': {
            'console_stderr': {
                # Sends log messages with log level ERROR or higher to stderr
                'class': 'logging.StreamHandler',
                'level': 'ERROR',
                'formatter': 'include_process',
                'stream': sys.stderr
            },
            'console_stdout': {
                # Sends log messages with log level lower than ERROR to stdout
                'class': 'logging.StreamHandler',
                'level': 'DEBUG',
                'formatter': 'include_process',
                'filters': ['exclude_errors'],
                'stream': sys.stdout
            },
            'file': {
                # Sends all log messages to a file
                'class': 'logging.FileHandler',
                'level': 'DEBUG',
                'formatter': 'include_process',
                'filename': 'uploader.log',
                'encoding': 'utf8'
            }
        },
        'root': {
            # In general, this should be kept at 'NOTSET'.
            # Otherwise it would interfere with the log levels set for each handler.
            'level': 'NOTSET',
            'handlers': ['console_stderr', 'console_stdout', 'file']
        },
    }

    def get_log_file(self):
        return self.config["handlers"]["file"]["filename"]

    def set_log_file(self, log_file_path):
        self.config["handlers"]["file"]["filename"] = log_file_path

    def initialize(self):
        logging.config.dictConfig(self.config)


class UploaderException(BaseException):
    pass


class FolderAlreadyExistsException(UploaderException):
    pass


class NotValidFolderException(UploaderException):
    pass


def parse_and_validate_jwt_token(jwt_token, public_key_path):
    # type: (str, str) -> dict

    if not re.match('^(eyJhbGci[a-zA-Z0-9_-]+)\\.([a-zA-Z0-9_-]+)\\.([a-zA-Z0-9_-]+)$', jwt_token):
        raise UploaderException('Provide correct JWT token')

    with open(public_key_path, "rb") as key_file:
        public_key = serialization.load_pem_public_key(
            key_file.read(),
            backend=default_backend()
        )

    try:
        return jwt.decode(jwt_token, public_key)
    except Exception:
        raise UploaderException('JWT token has not been verified: {}'.format(jwt_token))


def stoppable_sleep(wait_time, wake_up_condition, interval=5):
    # type: (float, callable, int) -> None

    logging.info("Waiting for {} seconds".format(wait_time))

    for _ in range(0, int(wait_time / interval)):
        if wake_up_condition():
            break

        time.sleep(interval)


def deep_dict_merge(source, destination):
    """
    Merge two dictionaries into one.
    """

    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            deep_dict_merge(value, node)
        else:
            destination[key] = value

    return destination



#!/usr/bin/env python

import csv
import logging
import os
import re
import tarfile
from abc import ABCMeta, abstractmethod

from soyuz.dx.variables import Type


class DataFile(object):
    __metaclass__ = ABCMeta

    def __init__(self, location, name, seq_folder):
        self.__location = location
        self.__name = name
        self.__seq_folder = seq_folder

    def get_location(self):
        return self.__location

    def get_name(self):
        return self.__name

    @property
    def is_valid(self):
        return bool(self.get_regex().match(self.get_name()))

    def get_seq_folder_name(self):
        return self.__seq_folder.get_name()

    def get_seq_folder_path(self):
        return self.__seq_folder.get_path()

    def get_relative_path(self):
        relpath = os.path.relpath(self.get_location(), self.__seq_folder.get_path())
        if relpath == ".":
            return ""
        return relpath

    def get_full_path(self):
        return os.path.join(self.__location, self.__name)

    @abstractmethod
    def get_type(self):
        raise NotImplementedError()

    @abstractmethod
    def get_regex(self):
        raise NotImplementedError()

    @abstractmethod
    def get_sample_id(self):
        raise NotImplementedError()


class RawDataFile(DataFile):
    def get_sample_id(self):
        return None

    def get_type(self):
        return None

    def get_regex(self):
        return re.compile(".+?")


class NgsDataFile(DataFile):
    def get_sample_id(self):
        return None

    def get_type(self):
        return "bam"

    def get_regex(self):
        # We filter out files which starts by __, because it's a temp files which causing
        # the issue when tar files are being uncompressed in Demux step, examples:
        #  __0001.bcl.bgzf.tmp.bci
        #  __0001.bcl.tmp
        return re.compile(".*Data/Intensities/(BaseCalls/)?L.*/(?!__).*")

    @property
    def is_valid(self):
        return bool(self.get_regex().match(self.get_full_path()))


class TarDataFile(DataFile):
    def __init__(self, location, name, seq_folder, data_file_type):
        super(TarDataFile, self).__init__(location, name, seq_folder)

        self.__files = []
        self.__data_file_type = data_file_type
        self.__tar_name = name
        self.__tar_location = location
        self.__tar_full_path = os.path.join(self.__tar_location, self.__tar_name)

    def get_sample_id(self):
        return None

    def get_type(self):
        return self.__data_file_type

    def get_regex(self):
        return re.compile(".+?")

    def get_relative_path(self):
        return ''

    def add_files(self, list_of_data_file_path):
        self.__files += list_of_data_file_path

    def add_file(self, data_file_path):
        self.__files.append(data_file_path)

    def get_files(self):
        return self.__files

    def create_tar_archive(self):
        logging.info('Creating tar archive: {}'.format(self.__tar_name))

        if not os.path.exists(self.__tar_location):
            os.makedirs(self.__tar_location)

        archive = tarfile.open(self.__tar_full_path, 'w:gz')
        for f in self.__files:
            f_rel = os.path.relpath(f, self.get_seq_folder_path())
            logging.info('>>> Adding {} to archive'.format(f_rel))
            archive.add(f, arcname=f_rel)

        archive.close()


class WesDataFile(DataFile):
    def get_regex(self):
        return re.compile("(?P<sample_id>[A-Za-z0-9-.]*)_[A-Za-z0-9-_]*.(?P<extension>fastq.gz|fastq|bam)")

    def get_type(self):
        m = self.get_regex().match(self.get_name())
        if m:
            extension = m.group("extension")
            if extension == "fastq.gz" or extension == "fastq":
                return Type.FASTQ
            if extension == "bam":
                return Type.BAM
        return None

    def get_sample_id(self):
        m = self.get_regex().match(self.get_name())
        if m:
            return m.group("sample_id")
        return None


class PersonalisDataFile(DataFile):
    def get_regex(self):
        return re.compile(
            "[A-Za-z0-9-]*_(?P<sample_id>[A-Za-z0-9-.]*)_[A-Za-z0-9-_.]*.(?P<extension>fastq.gz|fastq|bam)")

    def get_type(self):
        m = self.get_regex().match(self.get_name())
        if m:
            extension = m.group("extension")
            if extension == "fastq.gz" or extension == "fastq":
                return Type.FASTQ
            if extension == "bam":
                return Type.BAM
        return None

    def get_sample_id(self):
        m = self.get_regex().match(self.get_name())
        if m:
            return m.group("sample_id")
        return None


class QcDataFile(DataFile):
    @property
    def is_valid(self):
        if not super(QcDataFile, self).is_valid:
            return False

        if self.get_type() == Type.CSV:
            sniffer = csv.Sniffer()
            with open(self.get_full_path(), 'r') as f:
                data = f.read()
            if sniffer.sniff(data).delimiter != ',':
                logging.error("CSV file {}: expected a comma delimiter, but actual is {}.".format(self.get_full_path(),
                                                                                                  sniffer.sniff(
                                                                                                      data).delimiter))
                return False
            if not sniffer.has_header(data):
                logging.error("CSV file {} doesn't have a header.".format(self.get_full_path()))
                return False
            with open(self.get_full_path(), 'r') as f:
                for line in csv.DictReader(f):
                    if not re.compile("^[a-zA-Z0-9\\-_.]+$").match(line['Natera Sample ID']):
                        logging.error(
                            "CSV file {} doesn't have a valid 'Natera Sample ID'.".format(self.get_full_path()))
                        return False
                    if not line['aveDedupOnTarget']:
                        logging.error("CSV file {} doesn't have an 'aveDedupOnTarget'.".format(self.get_full_path()))
                        return False
                    metrics = ["Total recovery (ug)", "%gDNA", "A260/A280", "Library into Capture (ng)",
                               "PreCap BioA size (bp)", "Post capture Library concentration (ng/uL)",
                               "PostCap BioA size (bp)", "totReads", "totPairedReads", "averageBaseQ", "combinedQ30pct",
                               "r1Q30pct", "r2Q30pct", "mapq", "pctAlignedToGenome", "gcPCT", "pctDuplicate",
                               "pctOnPanel"]
                    for key, val in line.items():
                        if key in metrics and not val:
                            logging.warning(
                                "CSV file {} doesn't have a metric {} defined.".format(self.get_full_path(), key))
                    if line["% Tumor Fraction"] and not line["ffpe100x"]:
                        logging.warning("CSV file {} doesn't have 'ffpe100x' defined.".format(self.get_full_path()))
                    if not line["% Tumor Fraction"] and not line["norm30x"]:
                        logging.warning("CSV file {} doesn't have 'norm30x' defined.".format(self.get_full_path()))
        return True

    def get_regex(self):
        return re.compile("[A-Za-z0-9-_ ]*.(?P<extension>csv|pdf|xlsx)")

    def get_type(self):
        m = self.get_regex().match(self.get_name())
        if m:
            extension = m.group("extension")
            if extension == "csv":
                return Type.CSV
            if extension == "pdf":
                return Type.PDF
            if extension == "xlsx":
                return Type.XLSX
        return None

    def get_sample_id(self):
        return None


class PersonalisQcDataFile(QcDataFile):
    @property
    def is_valid(self):
        if not super(QcDataFile, self).is_valid:
            return False

        if self.get_type() == Type.CSV:
            sniffer = csv.Sniffer()
            with open(self.get_full_path(), 'r') as f:
                data = f.read()
            if sniffer.sniff(data).delimiter != ',':
                logging.error("CSV file {}: expected a comma delimiter, but actual is {}.".format(self.get_full_path(),
                                                                                                  sniffer.sniff(
                                                                                                      data).delimiter))
                return False
            if not sniffer.has_header(data):
                logging.error("CSV file {} doesn't have a header.".format(self.get_full_path()))
                return False
            with open(self.get_full_path(), 'r') as f:
                for line in csv.DictReader(f):
                    if not re.compile("^[a-zA-Z0-9\\-_.]+$").match(line['Natera Sample ID']):
                        logging.error(
                            "CSV file {} doesn't have a valid 'Natera Sample ID'.".format(self.get_full_path()))
                        return False
                    if not line['aveDedupOnTarget']:
                        logging.error("CSV file {} doesn't have an 'aveDedupOnTarget'.".format(self.get_full_path()))
                        return False
                    metrics = ["% Tumor Fraction", "extraction_ng", "DIN", "Q30_reads", "mapQ", "pct_reads_mapped",
                               "pct_GC", "duplication_rate", "cap_spec", "aveDedupOnTarget", "f_above_30x",
                               "f_above_100x"]
                    for key, val in line.items():
                        if key in metrics and not val:
                            logging.warning(
                                "CSV file {} doesn't have a metric {} defined.".format(self.get_full_path(), key))
        return True


class SangerDataFile(DataFile):
    def get_sample_id(self):
        return None

    def get_type(self):
        m = self.get_regex().match(self.get_name())
        if m:
            extension = m.group("extension")
            if extension == "ab1":
                return Type.AB1
            if extension == "fsa":
                return Type.FSA

    def get_regex(self):
        return re.compile(".+?(?P<extension>ab1|fsa)")



import os

import requests

from soyuz.utils import read_in_chunks


class ExodusApi(object):
    def __init__(self, exodus_url, exodus_username, exodus_password, chunk_size, exodus_aws_access_key_id,
                 exodus_aws_secret_access_key, exodus_s3_bucket_name, exodus_aws_region, exodus_sample_source):
        self._base_headers = {}
        self._exodus_url = exodus_url
        self._exodus_username = exodus_username
        self._exodus_password = exodus_password
        self._chunk_size = chunk_size

        if exodus_aws_access_key_id:
            self._base_headers['X-AWS_ACCESS_KEY_ID'] = exodus_aws_access_key_id

        if exodus_aws_secret_access_key:
            self._base_headers['X-AWS_SECRET_ACCESS_KEY'] = exodus_aws_secret_access_key

        if exodus_s3_bucket_name:
            self._base_headers['X-AWS_S3_BUCKET_NAME'] = exodus_s3_bucket_name

        if exodus_aws_region:
            self._base_headers['X-AWS_DEFAULT_REGION'] = exodus_aws_region

        if exodus_sample_source:
            self._base_headers['X-SAMPLE_SOURCE'] = exodus_sample_source

    def create_record(self, name, folder, types, properties, tags, details):
        data = {'name': name,
                'folder': folder,
                'types': types,
                'properties': properties,
                'tags': tags,
                'details': details
                }

        response = requests.post(os.path.join(self._exodus_url, "2.0/data/records/new"),
                                 json=data,
                                 auth=(self._exodus_username, self._exodus_password),
                                 headers=self._base_headers)

        response.raise_for_status()

        return response.json()

    def upload_local_file(self, file_full_path, name, folder, types, properties, tags):
        new_file_data = {'name': name, 'folder': folder, 'types': types, 'properties': properties, 'tags': tags}
        new_file_response = requests.post(os.path.join(self._exodus_url, "2.0/data/files/new"),
                                          json=new_file_data,
                                          auth=(self._exodus_username, self._exodus_password),
                                          headers=self._base_headers)
        new_file_response.raise_for_status()
        new_file_response = new_file_response.json()
        file_id = new_file_response['file']['link']
        upload_id = new_file_response['uploadId']

        with open(file_full_path, 'rb') as file_object:
            parts = []
            for i, chunk in enumerate(read_in_chunks(file_object, self._chunk_size)):
                part_upload_response = requests.post(os.path.join(self._exodus_url, "2.0/data/files/part-upload"),
                                                     json={'file': {'link': file_id}, 'uploadId': upload_id,
                                                           'part': i + 1},
                                                     auth=(self._exodus_username, self._exodus_password),
                                                     headers=self._base_headers)
                part_upload_response.raise_for_status()

                upload_url = part_upload_response.json()['uploadUrl']
                upload_response = requests.put(upload_url, data=chunk, headers=self._base_headers)
                upload_response.raise_for_status()

                parts.append({'eTag': upload_response.headers['ETag'], 'part': i + 1})

        finish_upload_response = requests.post(os.path.join(self._exodus_url, "2.0/data/files/finish-upload"),
                                               json={'file': {'link': file_id},
                                                     'uploadId': upload_id,
                                                     'contentLength': os.path.getsize(file_full_path),
                                                     'parts': parts},
                                               auth=(self._exodus_username, self._exodus_password),
                                               headers=self._base_headers)
        finish_upload_response.raise_for_status()

        return finish_upload_response.json()



#!/usr/bin/env python
import errno
import json
import logging
import os
import re
import threading
import time
from abc import ABCMeta, abstractmethod

import zope.event.classhandler

from soyuz.data.files import WesDataFile, RawDataFile, QcDataFile, PersonalisDataFile, PersonalisQcDataFile, \
    SangerDataFile, NgsDataFile, TarDataFile
from soyuz.dx.context import Context
from soyuz.dx.variables import Type
from soyuz.event.upload import UploadJobTerminated, SeqFolderUploadStart, SeqFolderUploadComplete, \
    SeqFolderUploadTerminated, DataFileUploadStart, DataFileUploadFinish
from soyuz.utils import UploaderException, NotValidFolderException, stoppable_sleep


class WatchDirectory(object):
    UPLOADED_FILE = ".uploaded"
    UPLOADING_FILE = ".uploading"

    def __init__(self, context, folder_path):
        # type: (Context, str) -> None
        self.__context = context
        self.__path = folder_path
        self.__lock = threading.Lock()

        self.__init_listeners()

    def is_uploaded(self, folder):
        return str(folder) in self.__read_file(self.__uploaded_file)

    def is_uploading(self, folder):
        return str(folder) in self.__read_file(self.__uploading_file)

    def complete(self, folder):
        self.remove_folder_from_uploading_file(str(folder))
        self.__write_file(self.__uploaded_file, str(folder))

    def start_upload(self, folder):
        self.remove_folder_from_uploaded_file(str(folder))
        self.__write_file(self.__uploading_file, str(folder))

    def remove_uploaded_file(self):
        if os.path.exists(self.__uploaded_file):
            os.remove(self.__uploaded_file)

    def remove_uploading_file(self):
        if os.path.exists(self.__uploading_file):
            os.remove(self.__uploading_file)

    def remove_folder_from_uploading_file(self, seq_folder_name):
        if os.path.exists(self.__uploading_file):
            self.__remove_from_file(self.__uploading_file, seq_folder_name)

    def remove_folder_from_uploaded_file(self, seq_folder_name):
        if os.path.exists(self.__uploaded_file):
            self.__remove_from_file(self.__uploaded_file, seq_folder_name)

    def __init_listeners(self):
        @zope.event.classhandler.handler(SeqFolderUploadStart)
        def on_seq_folder_upload_start(event):
            # type: (SeqFolderUploadStart) -> None

            self.start_upload(event.seq_folder.get_name())

        @zope.event.classhandler.handler(SeqFolderUploadComplete)
        def on_seq_folder_upload_complete(event):
            # type: (SeqFolderUploadComplete) -> None

            self.complete(event.seq_folder.get_name())

        @zope.event.classhandler.handler(SeqFolderUploadTerminated)
        def on_seq_folder_upload_terminated(event):
            # type: (SeqFolderUploadTerminated) -> None

            self.remove_folder_from_uploading_file(event.seq_folder.get_name())
            self.remove_folder_from_uploading_file(event.seq_folder.get_name())

    @property
    def __uploaded_file(self):
        return os.path.join(self.__path, self.UPLOADED_FILE)

    @property
    def __uploading_file(self):
        return os.path.join(self.__path, self.UPLOADING_FILE)

    def get_seq_folders(self):
        return self.__get_seq_folders(self.__path)

    def __get_seq_folders(self, root):
        result = []
        if not os.path.exists(root):
            logging.error("Watch directory {} doesn't exist".format(root))
            return result

        for f in os.listdir(root):
            path = os.path.join(root, f)
            if os.path.isdir(path):
                seq_folder = SeqFolderFactory.create(root, f, self.__context)
                if seq_folder:
                    result.append(seq_folder)
                else:
                    result += self.__get_seq_folders(path)
        return result

    @staticmethod
    def __read_file(file_with_history):
        result = []
        if not os.path.isfile(file_with_history):
            return result
        with open(file_with_history, "r") as f:
            for line in f.readlines():
                result.append(line.strip())
        return result

    def __write_file(self, file_with_history, seq_folder_name):
        try:
            self.__lock.acquire()
            with open(file_with_history, "a") as f:
                f.write("{}\n".format(seq_folder_name))
        finally:
            self.__lock.release()

    def __remove_from_file(self, file_with_history, seq_folder_name):
        try:
            self.__lock.acquire()
            with open(file_with_history, "r") as f:
                content = f.readlines()

            with open(file_with_history, "w") as f:
                for line in content:
                    if seq_folder_name in line:
                        continue

                    f.write(line)
        finally:
            self.__lock.release()


class SeqFolderBase(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, path):
        self._path = self.__get_abs_path(path).replace("\\", "/")
        self._upload_info_file_path = self.__resolve_upload_info_file_path(self.get_name())
        self._upload_info_update_lock = threading.Lock()

        with open(self._upload_info_file_path) as f:
            self._upload_info = json.load(f)

        self.__init_listeners()

    def get_name(self):
        return os.path.basename(self._path)

    def remove_upload_info_file(self):
        self._upload_info = {}
        with self._upload_info_update_lock:
            self._upload_info = {}
            with open(self._upload_info_file_path, "w") as f:
                f.write("{}")

    def get_path(self):
        return self._path

    @property
    @abstractmethod
    def is_valid(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_completed(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def complete(self):
        raise NotImplementedError()

    def is_older_than(self, seconds):
        return time.time() - os.path.getmtime(self.__str__()) > seconds

    @abstractmethod
    def list_files(self):
        raise NotImplementedError()

    def _get_uploaded_files_paths(self):
        # type: () -> list
        files_paths = []
        for file_path in self._upload_info:
            if self._upload_info[file_path].get('state', '') == 'uploaded':
                files_paths.append(file_path)

        return files_paths

    def _is_file_not_uploaded(self, path):
        return path not in self._upload_info or self._upload_info[path].get('state', '') != 'uploaded'

    def _on_data_file_upload_start(self, event):
        # type: (DataFileUploadStart) -> None
        data_file = event.data_file

        if data_file.get_seq_folder_name() == self.get_name():
            self._data_file_upload_start(data_file.get_full_path())

    def _on_data_file_upload_finish(self, event):
        # type: (DataFileUploadFinish) -> None
        data_file = event.data_file

        if data_file.get_seq_folder_name() == self.get_name():
            self._data_file_upload_finish(data_file.get_full_path())

    def _data_file_upload_start(self, data_file_path):
        # type: (str) -> None
        with self._upload_info_update_lock:
            self._upload_info[data_file_path] = {
                'mtime': os.path.getmtime(data_file_path),
                'state': 'upload_started',
                'upload_start_time': int(time.time())
            }

            with open(self._upload_info_file_path, 'w') as f:
                json.dump(self._upload_info, f)

    def _data_file_upload_finish(self, data_file_path):
        # type: (str) -> None
        with self._upload_info_update_lock:
            self._upload_info[data_file_path]['state'] = 'uploaded'
            self._upload_info[data_file_path]['upload_finish_time'] = int(time.time())

            with open(self._upload_info_file_path, 'w') as f:
                json.dump(self._upload_info, f)

    def _raise_exception_if_folder_does_not_exist(self):
        if not os.path.isdir(self._path):
            raise UploaderException("{} does not exist".format(self._path))

    @staticmethod
    def __get_abs_path(path):
        abs_path = os.path.abspath(os.path.expanduser(path))
        if not os.path.isdir(abs_path):
            raise UploaderException("{} does not exist".format(abs_path))
        return abs_path

    def __resolve_upload_info_file_path(self, seq_folder_name):
        upload_info_file_path = '/tmp/{}/upload.info'.format(seq_folder_name)

        if os.path.exists(upload_info_file_path):
            return upload_info_file_path

        path = os.path.dirname(upload_info_file_path)

        try:
            os.makedirs(path)
        except OSError as exc:  # Python ≥ 2.5
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise exc

        # create file with empty JSON
        with open(upload_info_file_path, "w") as f:
            f.write("{}")

        return upload_info_file_path

    def __init_listeners(self):
        @zope.event.classhandler.handler(DataFileUploadStart)
        def on_upload_job_terminated(event):
            # type: (DataFileUploadStart) -> None
            self._on_data_file_upload_start(event)

        @zope.event.classhandler.handler(DataFileUploadFinish)
        def on_new_upload_job(event):
            # type: (DataFileUploadFinish) -> None
            self._on_data_file_upload_finish(event)


class WesSignateraSeqFolder(SeqFolderBase):
    FOLDER_REGEX = re.compile("[A-Z0-9]*_[0-9]{8}")

    def __init__(self, path):
        SeqFolderBase.__init__(self, path)
        self.__wes_subfolder = os.path.join(self._path, "WES_data")
        self.__qc_subfolder = os.path.join(self._path, "QC_reports")

    @property
    def is_valid(self):
        if WesSignateraSeqFolder.FOLDER_REGEX.match(self.get_name()) is None \
                or not os.path.isdir(self.__wes_subfolder) \
                or not os.path.isdir(self.__qc_subfolder):
            return False

        uploaded_files = self._get_uploaded_files_paths()
        if len(uploaded_files) == 0 and len(self.__list_from_wes()) == 0:
            logging.info("{} does not contain any data".format(self.__wes_subfolder))
            return False
        if len(uploaded_files) == 0 and len(self.__list_from_qc_reports()) == 0:
            logging.info("{} does not contain any data".format(self.__qc_subfolder))
            return False

        success = True
        for f in self.list_files():
            if not f.is_valid:
                logging.info("{} has invalid format".format(f.get_full_path()))
                success = False
        return success

    def list_files(self):
        return self.__list_from_wes() + self.__list_from_qc_reports()

    def __list_from_wes(self):
        result = []
        for f in os.listdir(self.__wes_subfolder):
            data_file = WesDataFile(self.__wes_subfolder, f, self)
            if self._is_file_not_uploaded(data_file.get_full_path()):
                result.append(data_file)

        return result

    def __list_from_qc_reports(self):
        result = []
        for f in os.listdir(self.__qc_subfolder):
            data_file = QcDataFile(self.__qc_subfolder, f, self)
            if self._is_file_not_uploaded(data_file.get_full_path()):
                result.append(data_file)

        return result


class PersonalisSeqFolder(SeqFolderBase):
    FOLDER_REGEX = re.compile("^[A-Za-z0-9_-]{1,48}$")

    def __init__(self, path):
        SeqFolderBase.__init__(self, path)
        self.__alignment_subfolder = os.path.join(self._path, '{}', "Alignment")
        self.__fastq_subfolder = os.path.join(self._path, '{}', "FASTQ")
        self.__qc_metrics_path = self._path

    @property
    def is_valid(self):
        if PersonalisSeqFolder.FOLDER_REGEX.match(self.get_name()) is None:
            return False

        uploaded_files = self._get_uploaded_files_paths()
        if len(self.__list_qc_metrics(self._path)) == 0 and len(uploaded_files) == 0:
            logging.info("{} does not contain any data".format(self.__qc_metrics_path))
            return False

        for folder in os.listdir(self._path):
            if not os.path.isdir(os.path.join(self._path, folder)):
                continue
            if not os.path.isdir(self.__alignment_subfolder.format(folder)):
                return False
            if len(uploaded_files) == 0 and len(self.__list_from(self.__alignment_subfolder.format(folder))) == 0:
                logging.info("{} does not contain any data".format(self.__alignment_subfolder))
                return False
            if len(uploaded_files) == 0 and os.path.isdir(self.__fastq_subfolder.format(folder)) and \
                    len(self.__list_from(self.__fastq_subfolder.format(folder))) == 0:
                logging.info("{} does not contain any data".format(self.__fastq_subfolder))
                return False

        success = True
        for f in self.list_files():
            if not f.is_valid:
                logging.info("{} has invalid format".format(f.get_full_path()))
                success = False
        return success

    def list_files(self):
        list_files = self.__list_qc_metrics(self._path)
        for folder in os.listdir(self._path):
            if not os.path.isdir(os.path.join(self._path, folder)):
                continue
            list_files.extend(self.__list_from(self.__alignment_subfolder.format(folder)))
            if os.path.isdir(self.__fastq_subfolder.format(folder)):
                list_files.extend(self.__list_from(self.__fastq_subfolder.format(folder)))

        return list_files

    def __list_from(self, subfolder):
        result = []
        for f in os.listdir(subfolder):
            data_file = PersonalisDataFile(subfolder, f, self)
            if self._is_file_not_uploaded(data_file.get_full_path()):
                result.append(data_file)

        return result

    def __list_qc_metrics(self, path):
        result = []
        for f in os.listdir(path):
            if os.path.isfile(os.path.join(path, f)):
                data_file = PersonalisQcDataFile(path, f, self)
                if self._is_file_not_uploaded(data_file.get_full_path()):
                    result.append(data_file)

        return result


class RawSeqFolder(SeqFolderBase):
    def __init__(self, path):
        super(RawSeqFolder, self).__init__(path)

    @property
    def is_valid(self):
        return True

    def list_files(self):
        self._raise_exception_if_folder_does_not_exist()

        result = []
        for root, subdirs, files in os.walk(self._path):
            for f in files:
                data_file = RawDataFile(root, f, self)
                if self._is_file_not_uploaded(data_file.get_full_path()):
                    result.append(data_file)

        return result


class BgiSeqFolder(SeqFolderBase):
    FOLDER_REGEX = re.compile("V[A-Za-z0-9]{9}")

    def __init__(self, path):
        super(BgiSeqFolder, self).__init__(path)
        self.__lane_folders = [os.path.join(self._path, lane) for lane in ["L01", "L02", "L03", "L04"]]

    @property
    def is_completed(self):
        if self.is_valid:
            return all([os.path.isfile(os.path.join(lane_folder,
                                                    "{}_{}_FileInfo.csv".format(os.path.basename(self._path),
                                                                                os.path.basename(lane_folder)))) for
                        lane_folder in self.__lane_folders])
        return False

    @property
    def is_valid(self):
        if (BgiSeqFolder.FOLDER_REGEX.match(self.get_name())
                and all([os.path.join(self._path, folder) in self.__lane_folders for folder in os.listdir(self._path) if
                         os.path.isdir(os.path.join(self._path, folder))])):
            return True
        return False

    def list_files(self):
        self._raise_exception_if_folder_does_not_exist()

        result = []
        for root, subdirs, files in os.walk(self._path):
            for f in files:
                data_file = RawDataFile(root, f, self)
                if self._is_file_not_uploaded(data_file.get_full_path()):
                    result.append(data_file)

        return result


class SangerSeqFolder(SeqFolderBase):
    FOLDER_REGEX = re.compile("^[0-9\\-]{10}_[A-Za-z0-9_\\-]*_[A-Za-z0-9\\-]*_Q[0-9]*$")
    WITH_QUADRANT_FILES_COUNT = 96

    def __init__(self, path, check_interval):
        super(SangerSeqFolder, self).__init__(path)
        self.__check_interval = check_interval
        self.__force_stop = False

        @zope.event.classhandler.handler(UploadJobTerminated)
        def on_upload_job_terminated(event):
            # type: (UploadJobTerminated) -> None
            if self.get_name() == event.seq_folder_name:
                self.__force_stop = True

    @property
    def is_completed(self):
        return len(self.__list_files()) == self.WITH_QUADRANT_FILES_COUNT

    @property
    def is_valid(self):
        return bool(SangerSeqFolder.FOLDER_REGEX.match(self.get_name()))

    def list_files(self):
        while not self.__force_stop:
            files_for_upload = self.__list_files()
            number_of_uploaded_files = self._get_uploaded_files_paths()

            logging.info("Found {} files for {}".format(len(files_for_upload), self.get_name()))

            if len(files_for_upload) + len(number_of_uploaded_files) > self.WITH_QUADRANT_FILES_COUNT:
                raise NotValidFolderException('Data folder is not in a valid state')

            if len(files_for_upload) + len(number_of_uploaded_files) == self.WITH_QUADRANT_FILES_COUNT:
                return files_for_upload

            stoppable_sleep(self.__check_interval, lambda: self.__force_stop)

        return []

    def __list_files(self):
        self._raise_exception_if_folder_does_not_exist()

        result = []
        for root, subdirs, files in os.walk(self._path):
            for f in files:
                data_file = SangerDataFile(root, f, self)
                if self._is_file_not_uploaded(data_file.get_full_path()) and data_file.is_valid:
                    result.append(data_file)

        return result


class NgsSeqFolder(SeqFolderBase):
    FOLDER_REGEX = re.compile('^[0-9]{6,8}_[A-Za-z0-9]*_[0-9]*_[^_ ]*$')

    def __init__(self, path, completion_marker, check_interval, file_age_time, archive_temp_folder):
        super(NgsSeqFolder, self).__init__(path)

        self.__archive_part_index = 0
        self.__completion_marker = completion_marker
        self.__check_interval = check_interval
        self.__file_age_time = file_age_time
        self.__temp_tar_file_folder = os.path.join(archive_temp_folder, self.get_name())
        self.__force_stop = False

        @zope.event.classhandler.handler(UploadJobTerminated)
        def on_upload_job_terminated(event):
            # type: (UploadJobTerminated) -> None
            if self.get_name() == event.seq_folder_name:
                self.__force_stop = True

    @property
    def is_valid(self):
        return bool(NgsSeqFolder.FOLDER_REGEX.match(self.get_name()))

    def list_files(self):
        while not self.__force_stop:
            is_completion_marker_exists = os.path.exists(os.path.join(self._path, self.__completion_marker))
            wait_time = max(self.__check_interval,
                            self.__file_age_time) if is_completion_marker_exists else self.__check_interval

            stoppable_sleep(wait_time, lambda: self.__force_stop)

            files_to_upload = self.__find_files_to_upload()
            if len(files_to_upload) != 0:
                tar_archive_name = "upload_{:03d}.tar.gz".format(self.__archive_part_index)
                tar_archive = TarDataFile(self.__temp_tar_file_folder, tar_archive_name, self, Type.BCL)
                tar_archive.add_files(files_to_upload)
                tar_archive.create_tar_archive()

                self.__archive_part_index += 1

                yield tar_archive
            else:
                logging.info('List of files for upload is empty')

            if is_completion_marker_exists:
                logging.info('The completion marker {} was found'.format(self.__completion_marker))
                break

        if not self.__force_stop:
            yield self.__create_metadata_tar_archive()

    def _on_data_file_upload_start(self, event):
        # type: (DataFileUploadStart) -> None
        data_file = event.data_file

        if data_file.get_seq_folder_name() == self.get_name():
            # NGS works only with TarDataFile
            for data_file_full_path in data_file.get_files():
                self._data_file_upload_start(data_file_full_path)

    def _on_data_file_upload_finish(self, event):
        # type: (DataFileUploadFinish) -> None
        data_file = event.data_file

        if data_file.get_seq_folder_name() == self.get_name():
            # NGS works only with TarDataFile
            for data_file_full_path in data_file.get_files():
                self._data_file_upload_finish(data_file_full_path)

    def __find_files_to_upload(self):
        self._raise_exception_if_folder_does_not_exist()

        files_to_upload = []
        for root, subdirs, files in os.walk(self._path):
            for f in files:
                data_file = NgsDataFile(root, f, self)
                data_file_path = data_file.get_full_path()

                if data_file.is_valid:
                    cur_time = int(time.time())
                    cur_mtime = os.path.getmtime(data_file_path)
                    file_was_change = cur_mtime > self._upload_info.get(data_file_path, {}).get('mtime', 0)

                    if cur_time - cur_mtime > self.__file_age_time:
                        if file_was_change or self._is_file_not_uploaded(data_file_path):
                            files_to_upload.append(data_file_path)
                    else:
                        logging.info(
                            "Age of file {} is {} less then {}. Skipping...".format(data_file_path,
                                                                                    cur_time - cur_mtime,
                                                                                    self.__file_age_time))
        return files_to_upload

    def __create_metadata_tar_archive(self):
        metadata_files = ["RunInfo.xml", "Data/Intensities/config.xml", "runParameters.xml", "RunParameters.xml"]
        metadata_folders = ["InterOp"]
        metadata_tar_archive = TarDataFile(self.__temp_tar_file_folder, 'metadata.tar.gz', self, Type.METADATA)

        for metadata_file in metadata_files + metadata_folders:
            full_path = os.path.join(self._path, metadata_file)

            if os.path.exists(full_path):
                if self._is_file_not_uploaded(full_path):
                    metadata_tar_archive.add_file(full_path)

        metadata_tar_archive.create_tar_archive()

        return metadata_tar_archive


class SeqFolderFactory(object):
    @staticmethod
    def create(base_path, name, context):
        if BgiSeqFolder.FOLDER_REGEX.match(name):
            return BgiSeqFolder(os.path.join(base_path, name))
        if WesSignateraSeqFolder.FOLDER_REGEX.match(name):
            return WesSignateraSeqFolder(os.path.join(base_path, name))
        if SangerSeqFolder.FOLDER_REGEX.match(name):
            return SangerSeqFolder(os.path.join(base_path, name), context.get_sanger_check_interval())
        if NgsSeqFolder.FOLDER_REGEX.match(name):
            return NgsSeqFolder(os.path.join(base_path, name),
                                context.get_ngs_completion_marker(),
                                context.get_ngs_check_interval(),
                                context.get_ngs_file_age_time(),
                                context.get_ngs_archive_temp_folder())

        logging.error("Folder {} is not allowed.".format(os.path.join(base_path, name)))

        return None

    @staticmethod
    def create_from_full_path(full_path, context):
        base_path, name = SeqFolderFactory.__parse_full_path(full_path)
        return SeqFolderFactory.create(base_path, name, context)

    @staticmethod
    def __parse_full_path(full_path):
        path = full_path.rstrip('/')
        return os.path.dirname(path), os.path.basename(path)



import json
import logging

import requests

from soyuz.utils import UploaderException


class ConstellationApi(object):
    def __init__(self, token, constellation_url):
        # type: (str, str) -> None

        self.__token = token
        self.__url = constellation_url

    def start_upload_job(self, run_folder_name, site_id):
        # type: (str, str) -> str

        logging.info('Creating upload job for {}'.format(run_folder_name))

        try:
            headers = {'content-type': 'application/json', 'Authorization': 'Bearer {}'.format(self.__token)}
            data = json.dumps({'runFolder': run_folder_name, 'siteId': site_id}) if site_id else json.dumps(
                {'runFolder': run_folder_name})

            r = requests.post(url='{}/2.0/jobs/sequencingUpload'.format(self.__url),
                              headers=headers,
                              data=data)

            logging.info('Upload job created: {}'.format(r.text))

            return r.json()['jobId']
        except Exception as e:
            logging.error('Error creating upload job for {} with error: \'{}\''.format(self.__url, e))

    def get_job_status_by_id(self, job_id):
        # type: (str) -> str

        headers = {'content-type': 'application/json', 'Authorization': 'Bearer {}'.format(self.__token)}

        return requests.get(
            url='{}/2.0/jobs/sequencingUpload/{}'.format(self.__url, job_id),
            headers=headers).json()['status']



#!/usr/bin/env python

import logging
import os
from abc import ABCMeta, abstractmethod

from soyuz import __version__ as version
from soyuz.api.dx import DxApi
from soyuz.api.exodus import ExodusApi
from soyuz.dx.variables import Type, Property
from soyuz.utils import UploaderException


class Record(object):
    __metaclass__ = ABCMeta

    def __init__(self, api, basedir, name):
        self._api = api
        self._id = None
        self._details = None
        self._tags = None
        self._types = None
        self._properties = None
        self._record = None
        self._basedir = basedir
        self._name = name

    @abstractmethod
    def get_id(self):
        raise NotImplementedError()

    @abstractmethod
    def close(self):
        raise NotImplementedError()

    @abstractmethod
    def get_details(self):
        raise NotImplementedError()

    @abstractmethod
    def add_tags(self, tags):
        raise NotImplementedError()

    @abstractmethod
    def set_details(self, details):
        raise NotImplementedError()

    @abstractmethod
    def get_properties(self):
        raise NotImplementedError()

    @abstractmethod
    def set_properties(self, properties):
        raise NotImplementedError()

    @abstractmethod
    def add_property(self, key, value):
        raise NotImplementedError()

    @abstractmethod
    def is_closed(self):
        raise NotImplementedError()


class DNAnexusRecord(Record):
    def __init__(self, api, basedir, name):
        super(DNAnexusRecord, self).__init__(api, basedir, name)

        existing_records = self._api.find_data_objects(classname="record",
                                                       folder=os.path.join(basedir, name).replace("\\", "/"),
                                                       typename=Type.UPLOAD_SENTINEL,
                                                       properties={Property.RUN_FOLDER: name},
                                                       return_handler=True)
        if not existing_records:
            self.__record = self._api.create_record(types=[Type.UPLOAD_SENTINEL],
                                                    folder=os.path.join(basedir, name).replace("\\", "/"),
                                                    name="{}_upload_sentinel".format(name),
                                                    properties={Property.RUN_FOLDER: name,
                                                                Property.VERSION: version},
                                                    parents=True)
        else:
            self.__record = existing_records[0]

    def get_id(self):
        return self.__record.get_id()

    def close(self):
        return self.__record.close()

    def is_closed(self):
        return self.__record.describe(fields={'state'})['state'] == 'closed'

    def get_details(self):
        return self.__record.get_details()

    def set_details(self, details):
        return self.__record.set_details(details)

    def get_properties(self):
        return self.__record.get_properties()

    def set_properties(self, properties):
        return self.__record.set_properties(properties)

    def add_property(self, key, value):
        properties = self.__record.get_properties()
        properties[key] = value
        self.set_properties(properties)

    def add_tags(self, tags):
        return self.__record.add_tags(tags)


class ExodusRecord(Record):
    def __init__(self, api, basedir, name):
        super(ExodusRecord, self).__init__(api, basedir, name)
        self.__id = None
        self.__record = None
        self.__name = name
        self.__basedir = basedir
        self.__details = {}
        self.__properties = {Property.RUN_FOLDER: self.__name, Property.VERSION: version}
        self.__tags = []
        self.__types = []

    def close(self):
        data = self._api.create_record(
            name=self.__name,
            folder=os.path.join(self.__basedir, self.__name).replace("\\", "/"),
            types=[Type.UPLOAD_SENTINEL],
            properties=self.__properties,
            tags=self.__tags,
            details=self.__details
        )

        self.__id = data['link']
        logging.info("Successfully created: {}".format(os.path.join(self.__basedir, self.__name)))

        return self.__id

    def is_closed(self):
        return False

    def get_id(self):
        return self.__id

    def get_details(self):
        return self.__details

    def set_details(self, details):
        self.__details = details

    def add_tags(self, tags):
        self.__tags.extend(tags)
        self.__tags = list(set(self.__tags))

    def get_properties(self):
        return self.__properties

    def set_properties(self, properties):
        self.__properties = dict(properties)

    def add_property(self, key, value):
        self.__properties[key] = value


class RecordFactory(object):
    @staticmethod
    def create(api, basedir, name):
        if isinstance(api, DxApi):
            return DNAnexusRecord(api, basedir, name)
        if isinstance(api, ExodusApi):
            return ExodusRecord(api, basedir, name)
        raise UploaderException("API which can work with {} was not found".format(name))



#!/usr/bin/env python
import logging
import os
import xml.etree.ElementTree as ET
from abc import ABCMeta, abstractmethod

from soyuz.data.records import RecordFactory, DNAnexusRecord, ExodusRecord
from soyuz.dx.variables import Type, Property
from soyuz.utils import get_instrument_type_from_name


class SentinelBase(object):
    __metaclass__ = ABCMeta

    DATA_KEY = "data"
    FASTQ_KEY = "fastq"
    METRICS_KEY = "run_metrics"
    DX_LINK_KEY = "$dnanexus_link"

    def __init__(self, api, basedir, name):
        self.record = RecordFactory.create(api, basedir, name)

    @abstractmethod
    def add_file(self, data_file, file_id):
        raise NotImplementedError()

    def add_link(self, file_id):
        if isinstance(self.record, DNAnexusRecord):
            return {SentinelBase.DX_LINK_KEY: file_id}
        elif isinstance(self.record, ExodusRecord):
            return file_id

    def get_id(self):
        return self.record.get_id()

    def is_closed(self):
        return self.record.is_closed()

    def close(self):
        self.record.close()

    def get_properties(self):
        return self.record.get_properties()

    def add_property(self, key, value):
        self.record.add_property(key, value)


class WesSignateraSentinel(SentinelBase):
    def add_file(self, data_file, file_id):
        details = self.record.get_details()
        sample_id = data_file.get_sample_id()
        if sample_id:
            self.record.add_tags([sample_id])
            self.__add_data_file_details(details, sample_id, self.add_link(file_id))
        else:
            self.__add_metrics_details(details, self.add_link(file_id))
        self.record.set_details(details)

    @staticmethod
    def __add_data_file_details(details, sample_id, file_link):
        if SentinelBase.DATA_KEY not in details:
            details[SentinelBase.DATA_KEY] = {}
        data_details = details[SentinelBase.DATA_KEY]
        if sample_id not in data_details:
            data_details[sample_id] = []
        data_details[sample_id].append(file_link)

    @staticmethod
    def __add_metrics_details(details, file_link):
        if SentinelBase.METRICS_KEY not in details:
            details[SentinelBase.METRICS_KEY] = []
        details[SentinelBase.METRICS_KEY].append(file_link)


class PersonalisSentinel(SentinelBase):
    def add_file(self, data_file, file_id):
        details = self.record.get_details()
        sample_id = data_file.get_sample_id()
        if sample_id:
            self.record.add_tags([sample_id])
            if data_file.get_type() == Type.BAM:
                self.__add_data_file_details(details, sample_id, self.add_link(file_id))
            elif data_file.get_type() == Type.FASTQ:
                self.__add_fastq_file_details(details, sample_id, self.add_link(file_id))
        else:
            self.__add_metrics_details(details, self.add_link(file_id))
        self.record.set_details(details)

    @staticmethod
    def __add_data_file_details(details, sample_id, file_link):
        if SentinelBase.DATA_KEY not in details:
            details[SentinelBase.DATA_KEY] = {}
        data_details = details[SentinelBase.DATA_KEY]
        if sample_id not in data_details:
            data_details[sample_id] = []
        data_details[sample_id].append(file_link)

    @staticmethod
    def __add_fastq_file_details(details, sample_id, file_link):
        if SentinelBase.FASTQ_KEY not in details:
            details[SentinelBase.FASTQ_KEY] = {}
        data_details = details[SentinelBase.FASTQ_KEY]
        if sample_id not in data_details:
            data_details[sample_id] = []
        data_details[sample_id].append(file_link)

    @staticmethod
    def __add_metrics_details(details, file_link):
        if SentinelBase.METRICS_KEY not in details:
            details[SentinelBase.METRICS_KEY] = []
        details[SentinelBase.METRICS_KEY].append(file_link)


class SangerSentinel(SentinelBase):
    def add_file(self, data_file, file_id):
        details = self.record.get_details()
        file_type = data_file.get_type()

        if file_type:
            if file_type not in details:
                details[file_type] = []
            details[file_type].append(self.add_link(file_id))

        self.record.set_details(details)


class NgsSentinel(SentinelBase):
    def __init__(self, api, basedir, name, site_id, expected_upload_time):
        super(NgsSentinel, self).__init__(api, basedir, name)

        if not self.is_closed():
            if site_id:
                self.record.add_property(Property.SITE_ID, site_id)

                describe = api.get_record(self.record.get_id()).describe(incl_properties=True)

                # Double check siteId property if it exists, to be sure that we will not have any problems with it later
                # in other system parts
                if describe['properties'][Property.SITE_ID] != site_id:
                    raise Exception("Sentinel record has incorrect siteId property: {}, but should be: {}".format(
                        describe['properties'][Property.SITE_ID], site_id))

            if expected_upload_time:
                self.record.add_property(Property.EXPECTED_UPLOAD_TIME, str(expected_upload_time))

    def add_file(self, data_file, file_id):
        details = self.record.get_details()

        file_type = data_file.get_type()

        if file_type:
            if file_type not in details:
                details[file_type] = []
            details[file_type].append(self.add_link(file_id))

            if file_type == 'metadata':
                self.record.add_property(Property.INSTRUMENT_TYPE,
                                         self.__retrieve_instrument_type(data_file.get_seq_folder_path()))

        self.record.set_details(details)

    @staticmethod
    def __retrieve_instrument_type(run_folder):
        try:
            run_info = ET.parse(os.path.join(run_folder, 'RunInfo.xml')).getroot()
            run = run_info.find('Run')
            instrument = run.find('Instrument')

            return get_instrument_type_from_name(instrument.text)
        except IOError as e:
            logging.error("Failed to read RunInfo.xml in {}: {}".format(run_folder, e))
        except AttributeError as e:
            logging.error("Failed to find instrument type in RunInfo.xml: {}".format(e))
        except Exception as e:
            logging.error("Parse error: {}".format(e))

        return None


class RawSentinel(SentinelBase):
    def add_file(self, data_file, file_id):
        pass



#!/usr/bin/env python

import logging
import os
import subprocess
from abc import ABCMeta, abstractmethod

from soyuz.utils import UploaderException, NotValidFolderException


class Storage(object):
    __metaclass__ = ABCMeta

    def __init__(self, api, context):
        self._api = api
        self._context = context

    @abstractmethod
    def validate_target_dir(self, folder):
        raise NotImplementedError()

    @abstractmethod
    def upload_file(self, data_file, remote_folder, types, properties):
        raise NotImplementedError()

    @abstractmethod
    def remove_file(self, data_file, remote_folder):
        raise NotImplementedError()

    @abstractmethod
    def remove_target_dir(self, folder):
        raise NotImplementedError()


class DNAnexusStorage(Storage):
    NAME = "dnanexus"

    def __init__(self, api, context):
        # type: (DxApi, Context) -> None
        super(DNAnexusStorage, self).__init__(api, context)

        self.basedir = self._context.get_base_dir()
        self.token = self._context.get_token()
        self.ua_path = self._context.get_ua_path()
        self.ua_parameters = self._context.get_ua_parameters()
        self.__project = self._context.get_project_id()

    def get_project_id(self):
        return self.__project

    def get_project(self):
        return self._api.get_project(self.get_project_id())

    def get_file_by_id(self, file_id):
        return self._api.get_file(file_id)

    def validate_target_dir(self, folder):
        if not folder.is_valid:
            raise NotValidFolderException(
                "{} is not valid".format(folder.get_name()))

    def upload_file(self, data_file, remote_folder, types, properties):
        logging.info("Uploading {} to {}".format(data_file.get_full_path(), remote_folder))
        dx_file = self._api.upload_local_file(data_file.get_full_path(),
                                              folder=remote_folder,
                                              keep_open=True,
                                              parents=True)
        if dx_file:
            dx_file.add_types(types)
            dx_file.set_properties(properties)
            dx_file.close()
        else:
            raise UploaderException("Failed to upload {}".format(data_file.get_full_path()))
        return dx_file.get_id()

    def remove_target_dir(self, folder):
        logging.debug("{} folder will be removed from DNANexus".format(folder.get_name()))

        project = self._api.get_project(self.get_project_id())
        project.remove_folder(os.path.join(self.basedir, folder.get_name()), recurse=True, force=True)

    def remove_file(self, data_file, remote_folder):
        files = self._api.find_data_objects(classname="file", name=data_file.get_name(), folder=remote_folder,
                                            return_handler=True)
        for remote_file in files:
            describe = remote_file.describe(fields={'state', 'name'})
            if describe['state'] != 'closed':
                logging.warning("{}{} file will be removed from DNANexus".format(remote_folder, describe['name']))
                remote_file.remove()


class DNAnexusStorageUA(DNAnexusStorage):
    def upload_file(self, data_file, remote_folder, types, properties):
        logging.info("Uploading {} to {}".format(data_file.get_full_path(), remote_folder))
        args = [r'"{}"'.format(self.ua_path), data_file.get_full_path().replace("(", "\(").replace(")", "\)")]
        args.extend(["--auth-token", self.token])
        args.extend(["-p", self.get_project_id()])
        args.extend(["-f", remote_folder])
        args.extend([self.ua_parameters])
        args.extend(["--type {}".format(_type) for _type in types])
        args.extend(["--property {}={}".format(key, val) for key, val in properties.items()])
        args.extend(["-g"])
        args.extend(["--do-not-compress"])
        file_id = subprocess.check_output(" ".join(args), shell=True).strip().decode('utf8').replace("'", '"')
        return file_id


class ExodusStorage(Storage):
    NAME = "stella"

    def __init__(self, api, context):
        super(ExodusStorage, self).__init__(api, context)

        if not context.get_exodus_url():
            raise UploaderException("Exodus URL was not specified")

    def validate_target_dir(self, folder):
        if not folder.is_valid:
            raise NotValidFolderException(
                "{} is not valid".format(folder.get_name()))

    def upload_file(self, data_file, remote_folder, types, properties):
        logging.info("Uploading {} to {}".format(data_file.get_full_path(), remote_folder))
        result = self._api.upload_local_file(file_full_path=data_file.get_full_path(),
                                             name=data_file.get_name(),
                                             folder=remote_folder,
                                             types=types,
                                             properties=properties,
                                             tags=[])

        logging.info("Successfully uploaded: {} to {}".format(data_file.get_full_path(), remote_folder))

        return result['link']

    def remove_target_dir(self, folder):
        raise NotImplementedError()

    def remove_file(self, data_file, remote_folder):
        logging.warning("Remove file not supported for exodus storage")


class StorageType(object):
    DNANEXUS = DNAnexusStorage.NAME
    EXODUS = ExodusStorage.NAME
    DEFAULT = DNANEXUS
    ALL = [DNANEXUS, EXODUS]


class StorageFactory(object):
    @staticmethod
    def create(api, context):
        # type: (DxApi|ExodusApi, Context) -> Storage

        if context.get_storage() == DNAnexusStorage.NAME:
            return DNAnexusStorageUA(api, context) if context.get_ua_path() else DNAnexusStorage(api, context)
        elif context.get_storage() == ExodusStorage.NAME:
            return ExodusStorage(api, context)
        raise UploaderException("Storage with name {} was not found".format(context.get_storage()))



#!/usr/bin/env python
import errno
import logging
import os
import tarfile
import threading
import time
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from multiprocessing.pool import ThreadPool

import zope.event
import zope.event.classhandler

from soyuz.api.dx import DxApi
from soyuz.api.exodus import ExodusApi
from soyuz.data.files import DataFile
from soyuz.data.folders import WesSignateraSeqFolder, PersonalisSeqFolder, RawSeqFolder, BgiSeqFolder, SeqFolderBase, \
    SangerSeqFolder, NgsSeqFolder
from soyuz.data.storages import StorageFactory
from soyuz.dx.context import Context
from soyuz.dx.sentinels import WesSignateraSentinel, RawSentinel, PersonalisSentinel, SangerSentinel, NgsSentinel
from soyuz.dx.variables import Type, Property
from soyuz.event.upload import SeqFolderUploadStart, SeqFolderUploadComplete, UploadJobTerminated, \
    SeqFolderUploadTerminated, NewUploadJob, DataFileUploadStart, DataFileUploadFinish
from soyuz.utils import UploaderException, timeit, get_expected_time, NotValidFolderException, \
    LOGGER_FORMAT


@contextmanager
def poolcontext(*args, **kwargs):
    pool = ThreadPool(*args, **kwargs)
    yield pool
    pool.terminate()


class WatchUploader(object):
    __metaclass__ = ABCMeta

    def __init__(self, context, api):
        # type: (Context, DxApi|ExodusApi) -> None

        self.__context = context
        self.__api = api

    def watch(self, watch_dir):
        if self.__context.is_force_upload():
            watch_dir.remove_uploaded_file()
            watch_dir.remove_uploading_file()

        while True:
            for seq_folder in watch_dir.get_seq_folders():
                if watch_dir.is_uploaded(seq_folder.get_name()):
                    logging.info("{} already uploaded. Skipping".format(seq_folder.get_name()))
                    continue

                if watch_dir.is_uploading(seq_folder.get_name()):
                    logging.info("{} is uploading. Skipping".format(seq_folder.get_name()))
                    continue

                threading.Thread(target=self.__upload, args=(seq_folder,), daemon=True).start()

            time.sleep(self.__context.get_interval())

    def __upload(self, seq_folder):
        try:
            uploader = DxUploaderFactory.create(self.__context, self.__api, seq_folder)
            uploader.upload(seq_folder)
        except Exception as e:
            logging.error("{}. Skipping".format(e))


class DxUploaderBase(object):
    __metaclass__ = ABCMeta

    def __init__(self, context, api):
        # type: (Context, DxApi|ExodusApi) -> None

        self._context = context
        self._api = api
        self._storage = StorageFactory.create(self._api, context)
        self.__terminated_seq_folder_names = set()
        self.__upload_conditions = {}
        self.__seq_folder_name_to_upload_job_id = {}

        self.__init_listeners()

    def __call__(self, data_file):
        return self.upload_file(data_file)

    @timeit
    def upload(self, seq_folder):
        seq_folder_name = seq_folder.get_name()

        with SeqFolderLogUploader(logging.root, seq_folder_name, self._storage, self._context.get_base_dir()):
            try:
                if self._context.is_force_upload():
                    logging.info("Force upload is enabled, removing folder {} before upload".format(seq_folder_name))
                    self.remove_target_dir(seq_folder)
                    seq_folder.remove_upload_info_file()

                self._validate_target_dir(seq_folder)

                logging.info("Starting upload for {}".format(seq_folder_name))
                sentinel = self._create_sentinel(seq_folder_name)

                if sentinel.is_closed():
                    logging.info("The {} folder is already uploaded".format(seq_folder_name))
                    return

                zope.event.notify(SeqFolderUploadStart(seq_folder, sentinel))

                if self._context.is_constellation_portal_mode():
                    event = threading.Event()
                    self.__upload_conditions[seq_folder_name] = event
                    logging.info(
                        "Constellation mode is enabled, waiting for the job to be started from Constellation Portal")
                    event.wait()

                    sentinel.add_property(Property.JOB_ID, self.__seq_folder_name_to_upload_job_id[seq_folder_name])

                if self._context.get_process_count() > 1:
                    with poolcontext(processes=self._context.get_process_count()) as pool:
                        results = pool.map(self, seq_folder.list_files())
                else:
                    results = [self.upload_file(data_file) for data_file in seq_folder.list_files()]

                if seq_folder_name not in self.__terminated_seq_folder_names:
                    for data_file, file_id in results:
                        sentinel.add_file(data_file, file_id)

                    sentinel.close()
                    logging.info("{} folder has been successfully uploaded".format(seq_folder_name))
                    zope.event.notify(SeqFolderUploadComplete(seq_folder))
                else:
                    zope.event.notify(SeqFolderUploadTerminated(seq_folder))
                    logging.info("Upload {} folder has been terminated".format(seq_folder_name))
                    self.__terminated_seq_folder_names.remove(seq_folder_name)
            except NotValidFolderException as e:
                logging.error("Upload {} folder has been failed due to validation error: {}".format(seq_folder_name, e))
                self.remove_target_dir(seq_folder)
                raise e
            except BaseException as e:
                logging.error(e, exc_info=True)
                raise e

    def upload_file(self, data_file):
        if data_file.get_seq_folder_name() not in self.__terminated_seq_folder_names:
            zope.event.notify(DataFileUploadStart(data_file))

            remote_folder = os.path.join(self._context.get_base_dir(),
                                         data_file.get_seq_folder_name(),
                                         data_file.get_relative_path()).replace("\\", "/")
            types = self._get_additional_types(data_file)
            types.append(Type.UPLOAD_DATA)
            properties = self._get_additional_properties(data_file, data_file.get_seq_folder_name())
            properties[Property.RUN_FOLDER] = data_file.get_seq_folder_name()
            try:
                file_id = self._storage.upload_file(data_file, remote_folder, types, properties)
            except BaseException as ex:
                self._storage.remove_file(data_file, remote_folder)
                raise ex

            zope.event.notify(DataFileUploadFinish(data_file, file_id))

            return data_file, file_id
        else:
            return data_file, None

    def remove_target_dir(self, seq_folder):
        self._storage.remove_target_dir(seq_folder)

    def _validate_target_dir(self, folder):
        self._storage.validate_target_dir(folder)

    @abstractmethod
    def _create_sentinel(self, seq_folder_name):
        raise NotImplementedError()

    @abstractmethod
    def _get_additional_types(self, data_file):
        raise NotImplementedError()

    @abstractmethod
    def _get_additional_properties(self, data_file, seq_folder_name):
        raise NotImplementedError()

    def __init_listeners(self):
        @zope.event.classhandler.handler(UploadJobTerminated)
        def on_upload_job_terminated(event):
            # type: (UploadJobTerminated) -> None
            self.__terminated_seq_folder_names.add(event.seq_folder_name)

        @zope.event.classhandler.handler(NewUploadJob)
        def on_new_upload_job(event):
            # type: (NewUploadJob) -> None
            if event.seq_folder_name in self.__upload_conditions:
                logging.info('Upload job was found: {}, resuming upload'.format(event.job_id))
                self.__seq_folder_name_to_upload_job_id[event.seq_folder_name] = event.job_id
                self.__upload_conditions[event.seq_folder_name].set()


class WesSignateraDxUploader(DxUploaderBase):
    SEQ_FOLDER_TYPE = WesSignateraSeqFolder

    def _create_sentinel(self, seq_folder_name):
        return WesSignateraSentinel(self._api, self._context.get_base_dir(), seq_folder_name)

    def _get_additional_types(self, data_file):
        types = []
        data_type = data_file.get_type()
        if data_type:
            types.append(data_type)
            if data_type == Type.CSV and data_file.get_name().startswith("WES-QCMetrics"):
                types.append(Type.WESQCREPORT)
        return types

    def _get_additional_properties(self, data_file, seq_folder_name):
        properties = {}
        if data_file.get_sample_id():
            properties[Property.SAMPLE_REFERENCE] = "{}/{}".format(seq_folder_name, data_file.get_sample_id())
        return properties


class PersonalisDxUploader(DxUploaderBase):
    SEQ_FOLDER_TYPE = PersonalisSeqFolder

    def _create_sentinel(self, seq_folder_name):
        return PersonalisSentinel(self._api, self._context.get_base_dir(), seq_folder_name)

    def _get_additional_types(self, data_file):
        types = []
        data_type = data_file.get_type()
        if data_type:
            types.append(data_type)
            if data_type == Type.CSV and data_file.get_name().startswith("QCMetrics"):
                types.append(Type.WESQCREPORT)
        return types

    def _get_additional_properties(self, data_file, seq_folder_name):
        properties = {}
        if data_file.get_sample_id():
            properties[Property.SAMPLE_REFERENCE] = "{}/{}".format(seq_folder_name, data_file.get_sample_id())
        return properties


class RawDxUploader(DxUploaderBase):
    SEQ_FOLDER_TYPE = RawSeqFolder

    def _create_sentinel(self, seq_folder_name):
        return RawSentinel(self._api, self._context.get_base_dir(), seq_folder_name)

    def _get_additional_types(self, data_file):
        return []

    def _get_additional_properties(self, data_file, seq_folder_name):
        return {}


class BgiDxUploader(DxUploaderBase):
    SEQ_FOLDER_TYPE = BgiSeqFolder

    def _create_sentinel(self, seq_folder_name):
        return RawSentinel(self._api, self._context.get_base_dir(), seq_folder_name)

    def _get_additional_types(self, data_file):
        return []

    def _get_additional_properties(self, data_file, seq_folder_name):
        return {}


class SangerDxUploader(DxUploaderBase):
    SEQ_FOLDER_TYPE = SangerSeqFolder

    def _create_sentinel(self, seq_folder_name):
        return SangerSentinel(self._api, self._context.get_base_dir(), seq_folder_name)

    def _get_additional_types(self, data_file):
        return []

    def _get_additional_properties(self, data_file, seq_folder_name):
        return {}


class NgsDxUploader(DxUploaderBase):
    SEQ_FOLDER_TYPE = NgsSeqFolder

    def upload_file(self, data_file):
        data_file, file_id = super(NgsDxUploader, self).upload_file(data_file)

        logging.info("Removing uploaded tar file {}".format(data_file.get_full_path()))
        os.remove(data_file.get_full_path())

        return data_file, file_id

    def _create_sentinel(self, seq_folder_name):
        return NgsSentinel(self._api,
                           self._context.get_base_dir(),
                           seq_folder_name,
                           self._context.get_site_id(),
                           get_expected_time('NgsDxUploader', 'upload'))

    def _get_additional_types(self, data_file):
        return []

    def _get_additional_properties(self, data_file, seq_folder_name):
        return {}


class DxUploaderFactory(object):
    @staticmethod
    def create(context, api, seq_folder):
        # type: (Context, DxApi|ExodusApi, SeqFolderBase) -> DxUploaderBase

        if isinstance(seq_folder, BgiDxUploader.SEQ_FOLDER_TYPE):
            return BgiDxUploader(context, api)
        elif isinstance(seq_folder, RawDxUploader.SEQ_FOLDER_TYPE):
            return RawDxUploader(context, api)
        elif isinstance(seq_folder, WesSignateraDxUploader.SEQ_FOLDER_TYPE):
            return WesSignateraDxUploader(context, api)
        elif isinstance(seq_folder, PersonalisDxUploader.SEQ_FOLDER_TYPE):
            return PersonalisDxUploader(context, api)
        elif isinstance(seq_folder, SangerDxUploader.SEQ_FOLDER_TYPE):
            return SangerDxUploader(context, api)
        elif isinstance(seq_folder, NgsDxUploader.SEQ_FOLDER_TYPE):
            return NgsDxUploader(context, api)
        raise UploaderException(
            "Uploader for the folder {} was not found".format(seq_folder.get_name()))


class SeqFolderLogUploader:
    def __init__(self, logger, seq_folder_name, remote_storage, remote_base_dir):
        self._logger = logger
        self._remote_storage = remote_storage
        self._remote_base_dir = remote_base_dir
        self._seq_folder_name = seq_folder_name
        self._log_file_path = self._resolve_log_file_path(self._seq_folder_name)

        self._handler = logging.FileHandler(self._log_file_path)
        self._handler.setFormatter(logging.Formatter(LOGGER_FORMAT))

    def __enter__(self):
        self._logger.addHandler(self._handler)

    def __exit__(self, et, ev, tb):
        self._logger.removeHandler(self._handler)
        self._handler.close()

        self._upload_log_file()

    def _upload_log_file(self):
        if os.path.exists(self._log_file_path):
            remote_folder = os.path.join(self._remote_base_dir, self._seq_folder_name)
            tar_full_path = os.path.join(os.path.dirname(self._log_file_path), 'service.tar.gz')

            with tarfile.open(tar_full_path, 'w:gz') as t:
                t.add(self._log_file_path, arcname=os.path.basename(self._log_file_path))

            types = [Type.SERVICE]
            properties = {Property.RUN_FOLDER: self._seq_folder_name}
            tar_data_file = DataFile(os.path.dirname(tar_full_path), os.path.basename(tar_full_path), None)
            self._remote_storage.upload_file(tar_data_file, remote_folder, types, properties)

            if os.path.exists(tar_full_path):
                os.remove(tar_full_path)

            if os.path.exists(self._log_file_path):
                os.remove(self._log_file_path)

    def _resolve_log_file_path(self, seq_folder_name):
        log_file_path = '/tmp/{}/folder_upload.log'.format(seq_folder_name)
        path = os.path.dirname(log_file_path)

        try:
            os.makedirs(path)
        except OSError as exc:  # Python ≥ 2.5
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise exc

        return log_file_path



import logging
import os
import signal
import subprocess
import threading
import time
from subprocess import Popen


class SequenceFolderWatchDog(object):
    def __init__(self, seq_folder_path, check_interval, timeout, target_pid=None):
        # type: (str, float, float, int) -> None

        self.__seq_folder_path = os.path.abspath(os.path.expanduser(seq_folder_path))
        self.__target_pid = target_pid if target_pid else os.getpid()
        self.__check_interval = check_interval
        self.__timeout = timeout

        self.__watch_dog_thread = threading.Thread(target=self.__watch_dog, args=(), daemon=True)

    def start(self):
        self.__watch_dog_thread.start()

    def __watch_dog(self):
        while True:
            if self.__call_with_timeout(["ls", self.__seq_folder_path], self.__timeout):
                time.sleep(self.__check_interval)
            else:
                logging.warning("The {} sequence folder is not available, will kill a main process...".format(
                    self.__seq_folder_path))
                os.kill(self.__target_pid, signal.SIGKILL)

    @staticmethod
    def __call_with_timeout(cmd, timeout):
        # type: (list, float) -> bool

        start = time.time()
        p = Popen(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)

        while time.time() - start < timeout:
            if p.poll() is not None:
                return True

            time.sleep(0.5)

        p.kill()

        return False



#!/usr/bin/env python


class Property(object):
    JOB_ID = "jobId"
    TERMINATED = "terminated"
    RUN_FOLDER = "runFolder"
    SITE_ID = "siteId"
    INSTRUMENT_TYPE = "instrumentType"
    VERSION = "version"
    PRODUCT = "product"
    SAMPLE_REFERENCE = "uploadSampleReference"
    EXPECTED_UPLOAD_TIME = "expectedUploadTime"
    SOYUZ_ID = "soyuz_id"
    HEART_BEAT = "heart_beat"


class Type(object):
    HEART_BEAT = "HeartBeat"
    UPLOAD_DATA = "UploadData"
    UPLOAD_MANIFEST = "UploadManifest"
    UPLOAD_SENTINEL = "UploadSentinel"
    UPLOAD_JOB = "UPLOAD"
    BAM = "bam"
    FASTQ = "fastq"
    CSV = "csv"
    PDF = "pdf"
    XLSX = "xlsx"
    WESQCREPORT = "WESQcReport"
    AB1 = "ab1"
    FSA = "fsa"
    METADATA = "metadata"
    SERVICE = "service"
    BCL = "bcl"


class State(object):
    RUNNING = "running"
    WAITING = "waiting_on_input"
    TERMINATED = "terminated"
    DONE = "done"
    OPEN = "open"
    CLOSING = "closing"
    CLOSED = "closed"



import dxpy

from soyuz.configuration import Parameters
from soyuz.configuration import Settings
from soyuz.data.storages import StorageType
from soyuz.utils import UploaderException, parse_and_validate_jwt_token


class Context(object):
    def __init__(self, params, settings):
        # type: (Parameters, Settings) -> None

        self.__params = params
        self.__settings = settings
        self.__settings_from_dnanexus_profile = settings

        self.__project_id = None
        self.__dnanexus_token = None

        if self.__settings.get_storage() == StorageType.DNANEXUS:
            self.__dnanexus_token = self.__resolve_dnanexus_token()
            if not self.__dnanexus_token:
                raise UploaderException("Token was not specified")

    def set_settings_from_dnanexus_profile(self, settings):
        # type: (Settings) -> None

        self.__settings_from_dnanexus_profile = settings

    def set_project_id(self, project_id):
        # type: (str) -> None
        self.__project_id = project_id

    def get_project_id(self):
        # type: () -> str
        return self.__project_id

    def get_token(self):
        # type: () -> str

        return self.__dnanexus_token

    def get_interval(self):
        # type: () -> float

        return self.__params.get_interval() or self.__settings_from_dnanexus_profile.get_interval() or self.__settings.get_interval()

    def is_force_upload(self):
        # type: () -> bool

        return self.__params.is_force_upload()

    def is_constellation_portal_mode(self):
        # type: () -> bool

        return self.__params.is_constellation_portal_mode()

    def get_upload_jobs_check_interval(self):
        # type: () -> float

        return self.__params.get_upload_jobs_check_interval()

    def get_site_id(self):
        # type: () -> str

        return self.__params.get_site_id()

    def get_ngs_completion_marker(self):
        # type: () -> str

        return self.__params.get_ngs_completion_marker()

    def get_ngs_check_interval(self):
        # type: () -> float

        return self.__params.get_ngs_check_interval()

    def get_ngs_file_age_time(self):
        # type: () -> float

        return self.__params.get_ngs_file_age_time()

    def get_ngs_archive_temp_folder(self):
        # type: () -> str

        return self.__params.get_ngs_archive_temp_folder()

    def get_sanger_check_interval(self):
        # type: () -> float

        return self.__params.get_sanger_check_interval()

    def get_base_dir(self):
        # type: () -> str

        return self.__settings_from_dnanexus_profile.get_base_dir() or self.__settings.get_base_dir() or '/data/seq'

    def get_storage(self):
        # type: () -> str

        return self.__settings.get_storage()

    def get_exodus_url(self):
        # type: () -> str

        return self.__settings_from_dnanexus_profile.get_exodus_url() or self.__settings.get_exodus_url()

    def get_exodus_username(self):
        # type: () -> str

        return self.__settings_from_dnanexus_profile.get_exodus_username() or self.__settings.get_exodus_username()

    def get_exodus_password(self):
        # type: () -> str

        return self.__settings_from_dnanexus_profile.get_exodus_password() or self.__settings.get_exodus_password()

    def get_exodus_aws_access_key_id(self):
        # type: () -> str

        return self.__settings_from_dnanexus_profile.get_exodus_aws_access_key_id() or self.__settings.get_exodus_aws_access_key_id()

    def get_exodus_aws_secret_access_key(self):
        # type: () -> str

        return self.__settings_from_dnanexus_profile.get_exodus_aws_secret_access_key() or self.__settings.get_exodus_aws_secret_access_key()

    def get_exodus_s3_bucket_name(self):
        # type: () -> str

        return self.__settings_from_dnanexus_profile.get_exodus_s3_bucket_name() or self.__settings.get_exodus_s3_bucket_name()

    def get_exodus_aws_region(self):
        # type: () -> str

        return self.__settings_from_dnanexus_profile.get_exodus_aws_region() or self.__settings.get_exodus_aws_region()

    def get_exodus_sample_source(self):
        # type: () -> str

        return self.__settings_from_dnanexus_profile.get_exodus_sample_source() or self.__settings.get_exodus_sample_source()

    def get_constellation_url(self):
        # type: () -> str

        return self.__settings_from_dnanexus_profile.get_constellation_url() or self.__settings.get_constellation_url()

    def get_ua_path(self):
        # type: () -> str

        return self.__settings_from_dnanexus_profile.get_ua_path() or self.__settings.get_ua_path()

    def get_ua_parameters(self):
        # type: () -> str

        return self.__settings_from_dnanexus_profile.get_ua_parameters() or self.__settings.get_ua_parameters()

    def get_chunk_size(self):
        # type: () -> int

        return self.__settings_from_dnanexus_profile.get_chunk_size() or self.__settings.get_chunk_size()

    def get_process_count(self):
        # type: () -> int

        return self.__settings_from_dnanexus_profile.get_process_count() or self.__settings.get_process_count()

    def get_soyuz_id(self):
        # type: () -> str

        return self.__settings_from_dnanexus_profile.get_soyuz_id() or self.__settings.get_soyuz_id() \
               or self.__params.get_soyuz_id()

    def get_heartbeat_record_update_interval(self):
        # type: () -> str

        return self.__settings_from_dnanexus_profile.get_heartbeat_record_update_interval() \
               or self.__settings.get_heartbeat_record_update_interval() \
               or self.__params.get_heartbeat_record_update_interval()

    def __resolve_dnanexus_token(self):
        # type: () -> str

        if self.__params.get_jwt_token() or self.__settings.get_jwt_token():
            jwt_token = self.__params.get_jwt_token() or self.__settings.get_jwt_token()
            public_key_path = self.__params.get_jwt_token_public_key_path() or self.__settings.get_jwt_token_public_key_path()

            if not public_key_path:
                raise UploaderException('Please provide correct path to JWT token public key')

            return parse_and_validate_jwt_token(jwt_token, public_key_path)['soyuzUploadToken']
        else:
            return self.__params.get_token() or self.__settings.get_token()




import threading
import time

from soyuz.api.dx import DxApi
from soyuz.dx.variables import Type, Property


class Heartbeat(object):
    REMOTE_FOLDER_PATH = '/soyuz_agents'

    def __init__(self, api, soyuz_id, update_interval):
        # type: (DxApi, str, float) -> None

        self.__api = api
        self.__soyuz_id = soyuz_id
        self.__update_interval = update_interval

        self.__heartbeat_thread = threading.Thread(target=self.__heartbeat, args=(), daemon=True)

    def start(self):
        self.__heartbeat_thread.start()

    def __heartbeat(self):
        while True:
            record = self.__find_or_create_heartbeat_record()
            record.set_properties({Property.HEART_BEAT: str(int(time.time()))})
            time.sleep(self.__update_interval)

    def __find_or_create_heartbeat_record(self):
        records = self.__api.find_data_objects(classname="record",
                                               folder=self.REMOTE_FOLDER_PATH,
                                               typename=Type.HEART_BEAT,
                                               name=self.__soyuz_id,
                                               properties={Property.SOYUZ_ID: self.__soyuz_id},
                                               return_handler=True)
        if len(records) != 0:
            return records[0]

        return self.__api.create_record(types=[Type.HEART_BEAT],
                                        folder=self.REMOTE_FOLDER_PATH,
                                        name=self.__soyuz_id,
                                        properties={Property.SOYUZ_ID: self.__soyuz_id},
                                        close=True,
                                        parents=True)



import logging

import dxpy

from soyuz.configuration import Settings
from soyuz.utils import UploaderException


class DxApi(object):
    def __init__(self, token):
        # type: (str) -> None
        dxpy.set_security_context({'auth_token_type': 'Bearer', 'auth_token': token})

        self.__project_id = self.__resolve_project_id()
        dxpy.set_project_context(self.__project_id)
        dxpy.set_workspace_id(self.__project_id)

    def get_current_project_id(self):
        return self.__project_id

    def get_project(self, project_id):
        return dxpy.DXProject(project_id)

    def get_record(self, record_id):
        return dxpy.DXRecord(record_id)

    def get_job(self, job_id):
        return dxpy.DXJob(job_id)

    def get_file(self, file_id):
        return dxpy.DXFile(file_id)

    def upload_local_file(self, filename=None, file=None, media_type=None, keep_open=False,
                          wait_on_close=False, use_existing_dxfile=None, show_progress=False,
                          write_buffer_size=None, multithread=True, **kwargs):
        return dxpy.upload_local_file(filename=filename, file=file, media_type=media_type, keep_open=keep_open,
                                      wait_on_close=wait_on_close, use_existing_dxfile=use_existing_dxfile,
                                      show_progress=show_progress, write_buffer_size=write_buffer_size,
                                      multithread=multithread, **kwargs)

    def find_projects(self, name=None, name_mode='exact', properties=None, tags=None,
                      level=None, describe=False, explicit_perms=None, region=None,
                      public=None, created_after=None, created_before=None, billed_to=None,
                      limit=None, return_handler=False, first_page_size=100, containsPHI=None, **kwargs):
        return list(dxpy.bindings.search.find_projects(
            name=name, name_mode=name_mode, properties=properties, tags=tags,
            level=level, describe=describe, explicit_perms=explicit_perms, region=region,
            public=public, created_after=created_after, created_before=created_before, billed_to=billed_to,
            limit=limit, return_handler=return_handler, first_page_size=first_page_size, containsPHI=containsPHI,
            **kwargs
        ))

    def find_executions(self, classname=None, launched_by=None, executable=None, project=None,
                        state=None, origin_job=None, parent_job=None, no_parent_job=False,
                        parent_analysis=None, no_parent_analysis=False, root_execution=None,
                        created_after=None, created_before=None, describe=False,
                        name=None, name_mode="exact", tags=None, properties=None, limit=None,
                        first_page_size=100, return_handler=False, include_subjobs=True,
                        **kwargs):
        return list(dxpy.find_executions(
            classname=classname, launched_by=launched_by, executable=executable, project=project,
            state=state, origin_job=origin_job, parent_job=parent_job, no_parent_job=no_parent_job,
            parent_analysis=parent_analysis, no_parent_analysis=no_parent_analysis,
            root_execution=root_execution, created_after=created_after, created_before=created_before,
            describe=describe, name=name, name_mode=name_mode, tags=tags, properties=properties, limit=limit,
            first_page_size=first_page_size, return_handler=return_handler, include_subjobs=include_subjobs,
            **kwargs))

    def find_data_objects(self, classname=None, state=None, visibility=None,
                          name=None, name_mode='exact', properties=None,
                          typename=None, tag=None, tags=None,
                          link=None, project=None, folder=None, recurse=None,
                          modified_after=None, modified_before=None,
                          created_after=None, created_before=None,
                          describe=False, limit=None, level=None, region=None,
                          return_handler=False, first_page_size=100,
                          **kwargs):
        return list(dxpy.bindings.find_data_objects(
            classname=classname, state=state, visibility=visibility, name=name, name_mode=name_mode,
            properties=properties, typename=typename, tag=tag, tags=tags, link=link, project=project, folder=folder,
            recurse=recurse, modified_after=modified_after, modified_before=modified_before,
            created_after=created_after, created_before=created_before, describe=describe, limit=limit, level=level,
            region=region, return_handler=return_handler, first_page_size=first_page_size, **kwargs))

    def create_record(self, details=None, **kwargs):
        return dxpy.new_dxrecord(details=details, **kwargs)

    def load_settings_from_dnanexus_profile(self):
        # type: () -> Settings
        external_profiles = self.find_data_objects(classname="record",
                                                   project=self.__project_id,
                                                   folder='/profile/sequencingUpload/',
                                                   name='config', recurse=False, return_handler=True)

        if len(external_profiles) > 1:
            logging.warning("More than 1 settings profile exists in /profile/sequencingUpload/ folder, skipping...")
        elif len(external_profiles) == 0:
            return Settings(settings_dict={})

        if len(external_profiles) == 1:
            logging.info("Found a settings profile in /profile/sequencingUpload/ folder, loading...")

            return Settings(settings_dict=external_profiles[0].describe(incl_details=True)['details'])

    def __resolve_project_id(self):
        # type: () -> str

        try:
            projects = list(self.find_projects(level='UPLOAD'))

            if len(projects) == 0 or len(projects) > 1:
                raise UploaderException("Auth Token must have access to exactly 1 project with UPLOAD permission.")

            return projects[0]['id']
        except dxpy.exceptions.InvalidAuthentication:
            raise UploaderException("Authorization token is invalid or expired")



import logging
import time
from threading import Thread

import zope.event.classhandler
from dxpy import DXRecord, DXJobFailureError

from soyuz.api.dx import DxApi
from soyuz.constellation.api import ConstellationApi
from soyuz.dx.context import Context
from soyuz.dx.sentinels import SentinelBase
from soyuz.dx.variables import Type, Property
from soyuz.event.upload import SeqFolderUploadStart, UploadJobTerminated, SeqFolderUploadTerminated, NewUploadJob, \
    SeqFolderUploadComplete


class UploadJobWatcher(object):
    SOYUZ_JOB_MARKER = 'soyuzJob'

    def __init__(self, context, api, constellation_api, check_interval=30):
        # type: (Context, DxApi, ConstellationApi, float) -> None

        self.__seq_folder_name_to_job_ids = {}

        self.__context = context
        self.__dx_api = api
        self.__constellation_api = constellation_api
        self.__check_interval = check_interval

        self.__stop = False
        self.__watcher_thread = Thread(target=self._check_jobs, args=())
        self.__watcher_thread.daemon = True

    def watch_job(self, seq_folder_name, job_id):
        # type: (str, str) -> None

        self.__seq_folder_name_to_job_ids[seq_folder_name] = job_id

    def start(self):
        self.__watcher_thread.start()

    def stop(self):
        self.__stop = True

    def wait_upload_job_completion(self, run_folder_name):
        # type: (str) -> None

        if run_folder_name in self.__seq_folder_name_to_job_ids:
            job_id = self.__seq_folder_name_to_job_ids[run_folder_name]
            logging.info('Waiting until upload job with id: {} will be finished'.format(job_id))

            try:
                self.__dx_api.get_job(job_id).wait_on_done(interval=10, timeout=60 * 5)
            except DXJobFailureError:
                logging.info(
                    'Upload job with id: {} has been terminated or reached timeout (5 min) while waiting for the job to finish.'.format(
                        job_id))

                self.cleanup_seq_folder(run_folder_name)

    def cleanup_seq_folder(self, seq_folder_name):
        # type: (str) -> None

        logging.info('Cleanup {} folder'.format(seq_folder_name))

        manifest = self._get_manifest(seq_folder_name)

        details = manifest.get_details()
        properties = manifest.get_properties()
        types = manifest.types
        manifest_name = manifest.name

        full_run_folder_name = "{}/{}".format(self.__context.get_base_dir(), details['runFolder'])

        self._remove_data_from_run_folder(full_run_folder_name)
        self._remove_marker_from_upload_job(properties['jobId'])

        self._create_new_manifest(details, properties, types, manifest_name, full_run_folder_name)

    def _get_manifest(self, seq_folder_name):
        # type: (str) -> DXRecord | None

        manifests = self.__dx_api.find_data_objects(classname="record",
                                                    typename=Type.UPLOAD_MANIFEST,
                                                    properties={Property.RUN_FOLDER: seq_folder_name},
                                                    describe=True,
                                                    project=self.__context.get_project_id(),
                                                    return_handler=True)

        if not manifests:
            return None

        return manifests[0]

    def _remove_data_from_run_folder(self, full_run_folder):
        # type: (str) -> None

        logging.debug("Removing data from {}".format(full_run_folder))

        dx_project = self.__dx_api.get_project(self.__context.get_project_id())
        dx_project.remove_folder(full_run_folder, recurse=True, force=True)
        dx_project.new_folder(full_run_folder, parents=True)

    def _remove_marker_from_upload_job(self, job_id):
        # type: (str) -> None

        logging.debug("Removing marker property from {}".format(job_id))

        self.__dx_api.get_job(job_id).set_properties({self.SOYUZ_JOB_MARKER: None})

    def _create_new_manifest(self, details, properties, types, manifest_name, full_run_folder_name):
        # type: (dict, dict, list, str, str) -> None

        properties[Property.TERMINATED] = "true"
        record = self.__dx_api.create_record(types=types,
                                             folder=full_run_folder_name,
                                             name=manifest_name,
                                             project=self.__context.get_project_id(),
                                             properties=properties,
                                             details=details).close()
        logging.debug("Manifest copied to {}".format(record))

    def _check_jobs(self):
        while not self.__stop:
            finished_seq_folder_names = set()

            for job in self._get_upload_jobs_in_project(self.__context.get_project_id()):
                job_id = job['id']
                state = job['describe']['state'].upper()
                seq_folder_name = job['describe']['properties']['runFolder']

                if state == 'DONE' and job_id in self.__seq_folder_name_to_job_ids.values():
                    finished_seq_folder_names.add(seq_folder_name)
                elif state == 'TERMINATED' and job_id in self.__seq_folder_name_to_job_ids.values():
                    finished_seq_folder_names.add(seq_folder_name)
                    zope.event.notify(UploadJobTerminated(seq_folder_name))
                elif state == 'WAITING_ON_INPUT' and seq_folder_name not in self.__seq_folder_name_to_job_ids:
                    self.__seq_folder_name_to_job_ids[seq_folder_name] = job_id
                    zope.event.notify(NewUploadJob(job_id, seq_folder_name))

            for seq_folder_name in finished_seq_folder_names:
                del self.__seq_folder_name_to_job_ids[seq_folder_name]

            time.sleep(self.__check_interval)

    def _get_upload_jobs_in_project(self, project_id):
        # type: (str) -> list

        return self.__dx_api.find_executions(classname='job',
                                             properties={Property.PRODUCT: 'UPLOAD'},
                                             project=project_id,
                                             describe=True)


class UploadJobManager(object):
    def __init__(self, constellation_api, job_watcher, is_constellation_portal_mode):
        # type: (ConstellationApi, UploadJobWatcher, bool) -> None

        self.__constellation_api = constellation_api
        self.__job_watcher = job_watcher
        self.__is_constellation_portal_mode = is_constellation_portal_mode

        self.__init_listeners()

    def start_watcher(self):
        self.__job_watcher.start()

    def _create_upload_job(self, seq_folder_name, sentinel):
        # type: (str, SentinelBase) -> None

        site_id = sentinel.get_properties().get(Property.SITE_ID, None)
        job_id = self.__constellation_api.start_upload_job(seq_folder_name, site_id)
        sentinel.add_property(Property.JOB_ID, job_id)

        self.__job_watcher.watch_job(seq_folder_name, job_id)

    def __init_listeners(self):
        @zope.event.classhandler.handler(SeqFolderUploadStart)
        def on_seq_folder_upload_start(event):
            # type: (SeqFolderUploadStart) -> None

            if not self.__is_constellation_portal_mode:
                # Wait 1 seconds, we need this for avoid problem with missing sentinel record
                time.sleep(1)
                self._create_upload_job(event.seq_folder.get_name(), event.sentinel)

        @zope.event.classhandler.handler(SeqFolderUploadComplete)
        def on_seq_folder_upload_complete(event):
            # type: (SeqFolderUploadComplete) -> None

            self.__job_watcher.wait_upload_job_completion(event.seq_folder.get_name())

        @zope.event.classhandler.handler(SeqFolderUploadTerminated)
        def on_seq_folder_upload_terminated(event):
            # type: (SeqFolderUploadTerminated) -> None

            self.__job_watcher.cleanup_seq_folder(event.seq_folder.get_name())



from soyuz.dx.sentinels import SentinelBase


class SeqFolderUploadStart(object):
    def __init__(self, seq_folder, sentinel):
        # type: (SeqFolderBase, SentinelBase) -> None
        self.seq_folder = seq_folder
        self.sentinel = sentinel


class SeqFolderUploadComplete(object):
    def __init__(self, seq_folder):
        # type: (SeqFolderBase) -> None

        self.seq_folder = seq_folder


class SeqFolderUploadTerminated(object):
    def __init__(self, seq_folder):
        # type: (SeqFolderBase) -> None

        self.seq_folder = seq_folder


class UploadJobTerminated(object):
    def __init__(self, seq_folder_name):
        # type: (str) -> None
        self.seq_folder_name = seq_folder_name


class NewUploadJob(object):
    def __init__(self, job_id, seq_folder_name):
        # type: (str, str) -> None
        self.job_id = job_id
        self.seq_folder_name = seq_folder_name


class DataFileUploadStart(object):
    def __init__(self, data_file):
        # type: (DataFile) -> None

        self.data_file = data_file


class DataFileUploadFinish(object):
    def __init__(self, data_file, file_id):
        # type: (DataFile, str) -> None

        self.data_file = data_file
        self.file_id = file_id



#!/usr/bin/env python

import logging

from soyuz.api.dx import DxApi
from soyuz.api.exodus import ExodusApi
from soyuz.configuration import Parameters
from soyuz.configuration import Settings
from soyuz.constellation.api import ConstellationApi
from soyuz.constellation.job_manager import UploadJobManager, UploadJobWatcher
from soyuz.data.daemons import Daemon
from soyuz.data.folders import WesSignateraSeqFolder, BgiSeqFolder, RawSeqFolder, PersonalisSeqFolder, WatchDirectory, \
    SangerSeqFolder, NgsSeqFolder
from soyuz.data.storages import StorageType
from soyuz.dx.context import Context
from soyuz.dx.uploaders import WesSignateraDxUploader, BgiDxUploader, RawDxUploader, PersonalisDxUploader, \
    WatchUploader, SangerDxUploader, NgsDxUploader
from soyuz.heartbeat import Heartbeat
from soyuz.utils import UploaderException, Logging
from soyuz.watchdog import SequenceFolderWatchDog


def execute(params, settings, context=None, api=None, constellation_api=None):
    if params.get_action() in [Parameters.UPLOAD_ACTION, Parameters.WATCH_ACTION]:
        folder = None
        uploader = None

        if params.is_with_upload_job() or params.is_constellation_portal_mode():
            if not context.get_constellation_url():
                raise UploaderException('Constellation API URL was not specified')

            UploadJobManager(constellation_api,
                             UploadJobWatcher(context, api, constellation_api,
                                              context.get_upload_jobs_check_interval()),
                             params.is_constellation_portal_mode()).start_watcher()

        SequenceFolderWatchDog(params.get_folder(),
                               params.get_watch_dog_check_interval(),
                               params.get_watch_dog_timeout()).start()

        if params.is_with_heartbeat():
            if not context.get_soyuz_id():
                raise UploaderException("--with-heartbeat requires Soyuz ID")
            elif context.get_storage() != StorageType.DNANEXUS:
                raise UploaderException("--with-heartbeat works only with DNANexus storage")
            else:
                Heartbeat(api, context.get_soyuz_id(), context.get_heartbeat_record_update_interval()).start()

        if params.get_action() == Parameters.UPLOAD_ACTION:
            if params.get_upload_type() == Parameters.WES_SIGNATERA_UPLOAD:
                uploader = WesSignateraDxUploader(context, api)
                folder = WesSignateraSeqFolder(params.get_folder())
            elif params.get_upload_type() == Parameters.BGI_UPLOAD:
                uploader = BgiDxUploader(context, api)
                folder = BgiSeqFolder(params.get_folder())
            elif params.get_upload_type() == Parameters.RAW_UPLOAD:
                uploader = RawDxUploader(context, api)
                folder = RawSeqFolder(params.get_folder())
            elif params.get_upload_type() == Parameters.PERSONALIS_UPLOAD:
                uploader = PersonalisDxUploader(context, api)
                folder = PersonalisSeqFolder(params.get_folder())
            elif params.get_upload_type() == Parameters.SANGER_UPLOAD:
                uploader = SangerDxUploader(context, api)
                folder = SangerSeqFolder(params.get_folder(), context.get_sanger_check_interval())
            elif params.get_upload_type() == Parameters.NGS_UPLOAD:
                uploader = NgsDxUploader(context, api)
                folder = NgsSeqFolder(params.get_folder(),
                                      context.get_ngs_completion_marker(),
                                      context.get_ngs_check_interval(),
                                      context.get_ngs_file_age_time(),
                                      context.get_ngs_archive_temp_folder())

            if not folder or not uploader:
                raise UploaderException("Incorrect upload type")

            if not folder.is_valid:
                raise UploaderException("Data folder is not in a valid state")

            uploader.upload(folder)

        elif params.get_action() == Parameters.WATCH_ACTION:
            uploader = WatchUploader(context, api)

            if params.get_watch_type() == Parameters.START_WATCH:
                if params.foreground():
                    uploader.watch(WatchDirectory(context, params.get_folder()))
                else:
                    Daemon(uploader.watch, WatchDirectory(context, params.get_folder())).start()
            elif params.get_watch_type() == Parameters.STOP_WATCH:
                Daemon().stop()
            elif params.get_watch_type() == Parameters.STATUS_WATCH:
                Daemon().status()

    elif params.get_action() == Parameters.CONFIG_ACTION:
        if params.get_config_action() == Parameters.GET_CONFIG:
            if params.get_config_parameter_key() == "storage":
                logging.info(settings.get_storage())
            elif params.get_config_parameter_key() == "token":
                logging.info(settings.get_token())
            elif params.get_config_parameter_key() == "stella_url":
                logging.info(settings.get_exodus_url())
            elif params.get_config_parameter_key() == "constellation_url":
                logging.info(settings.get_constellation_url())
            elif params.get_config_parameter_key() == "basedir":
                logging.info(settings.get_base_dir())
            elif params.get_config_parameter_key() == "interval":
                logging.info(settings.get_interval())
            elif params.get_config_parameter_key() == "ua_path":
                logging.info(settings.get_ua_path())
            elif params.get_config_parameter_key() == "ua_parameters":
                logging.info(settings.get_ua_parameters())
            elif params.get_config_parameter_key() == "process_count":
                logging.info(settings.get_process_count())
            elif params.get_config_parameter_key() == "chunk_size":
                logging.info(settings.get_chunk_size())
            elif params.get_config_parameter_key() == "jwt_token":
                logging.info(settings.get_jwt_token())
            elif params.get_config_parameter_key() == "jwt_token_public_key_path":
                logging.info(settings.get_jwt_token_public_key_path())
            elif params.get_config_parameter_key() == "constellation_url":
                logging.info(settings.get_constellation_url())

        elif params.get_config_action() == Parameters.SET_CONFIG:
            if params.get_config_parameter_key() == "storage":
                settings.set_storage(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "token":
                settings.set_token(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "stella_url":
                settings.set_exodus_url(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "stella_username":
                settings.set_exodus_username(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "stella_password":
                settings.set_exodus_password(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "stella_aws_access_key_id":
                settings.set_exodus_aws_access_key_id(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "stella_aws_secret_access_key":
                settings.set_exodus_aws_secret_access_key(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "stella_s3_bucket_name":
                settings.set_exodus_s3_bucket_name(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "stella_aws_region":
                settings.set_exodus_aws_region(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "stella_sample_source":
                settings.set_exodus_sample_source(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "constellation_url":
                settings.set_constellation_url(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "basedir":
                settings.set_basedir(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "interval":
                settings.set_interval(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "ua_path":
                settings.set_ua_path(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "ua_parameters":
                settings.set_ua_parameters(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "process_count":
                settings.set_process_count(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "chunk_size":
                settings.set_chunk_size(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "jwt_token":
                settings.set_jwt_token(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "jwt_token_public_key_path":
                settings.set_jwt_token_public_key_path(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "constellation_url":
                settings.set_constellation_url(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "soyuz_id":
                settings.set_soyuz_id(params.get_config_parameter_value())
            elif params.get_config_parameter_key() == "heartbeat_record_update_interval":
                settings.set_heartbeat_record_update_interval(params.get_config_parameter_value())
            else:
                raise UploaderException("There is no parameter '{}'".format(params.get_config_parameter_key()))
            settings.dump()
            logging.info("Set {} to {}".format(params.get_config_parameter_key(), params.get_config_parameter_value()))


def main(_params=None, _settings=None):
    try:
        params = _params if _params else Parameters()
        settings = _settings if _settings else Settings(full_path_to_settings_file=params.get_setting_file_path())
        Logging().initialize()

        if params.get_action() == Parameters.CONFIG_ACTION:
            execute(params, settings)
        else:
            api = None
            constellation_api = None
            context = Context(params, settings)

            if context.get_storage() == StorageType.DNANEXUS:
                api = DxApi(context.get_token())
                context.set_settings_from_dnanexus_profile(api.load_settings_from_dnanexus_profile())
                constellation_api = ConstellationApi(context.get_token(), context.get_constellation_url())
                context.set_project_id(api.get_current_project_id())
            elif context.get_storage() == StorageType.EXODUS:
                api = ExodusApi(exodus_url=context.get_exodus_url(),
                                exodus_username=context.get_exodus_username(),
                                exodus_password=context.get_exodus_password(),
                                chunk_size=context.get_chunk_size(),
                                exodus_aws_access_key_id=context.get_exodus_aws_access_key_id(),
                                exodus_aws_secret_access_key=context.get_exodus_aws_secret_access_key(),
                                exodus_s3_bucket_name=context.get_exodus_s3_bucket_name(),
                                exodus_aws_region=context.get_exodus_aws_region(),
                                exodus_sample_source=context.get_exodus_sample_source())

            execute(params, settings, context, api, constellation_api)
    except Exception as e:
        logging.error(e, exc_info=True)
        quit(1)


if __name__ == "__main__":
    main()
